import argparse, os, re
from datetime import datetime
from .validators import valid_action, valid_amount, valid_promo_duration, valid_date, valid_limit, valid_time, valid_price, valid_duration, valid_expiration, valid_schedule, valid_month, valid_path

ACTIONS = ['bot','discount','message','post','profile','promotion', 
'test'
]

CATEGORIES_DEFAULT = [
  "images",
  "galleries",
  "videos"
]
DEFAULT_MESSAGE = ":)"
DEFAULT_GREETING = "hi! thanks for subscribing :3 do you have any preferences?"
DISCOUNT_MAX_AMOUNT = 55
DISCOUNT_MIN_AMOUNT = 10
DISCOUNT_MAX_MONTHS = 7
DISCOUNT_MIN_MONTHS = 1
EXPIRATION_ALLOWED = [1,3,7,30,99]
IMAGE_DOWNLOAD_LIMIT = 6
IMAGE_UPLOAD_LIMIT = 5
MESSAGE_CHOICES = ["all", "recent", "favorite", "renew on"]
PRICE_MINIMUM = 3
UPLOAD_MAX_DURATION = 6*6 # increments of 10 minutes; 6 = 1 hr
DURATION_ALLOWED = [1,3,7,30,99]
LIMIT_ALLOWED = [0,1,2,3,4,5,6,7,8,9,10,20,30,40,50,60,70,80,90,100]
PROMOTION_EXPIRATION_ALLOWED = [int(i) for i in range(30)]
PROMOTION_EXPIRATION_ALLOWED.insert(0,0)
PROMOTION_OFFER_LIMIT = [0,1,2,3,4,5,6,7,8,9,10]
PROMOTION_DURATION_ALLOWED = ["1 day","3 days","7 days","14 days","1 month","3 months","6 months","12 months"]

# Paths
MOUNT_PATH = "/opt/onlysnarf"
DOWNLOAD_PATH = os.path.join(MOUNT_PATH, "downloads")
CONFIG_PATH = os.path.join(MOUNT_PATH, "config.conf")
GOOGLE_PATH = os.path.join(MOUNT_PATH, "google_creds.txt")
SECRET_PATH = os.path.join(MOUNT_PATH, "client_secrets.json")
USERS_PATH = os.path.join(MOUNT_PATH, "users.json")
PROFILE_PATH = os.path.join(MOUNT_PATH, "profile.json")

# from pathlib import Path
# Path(MOUNT_PATH).mkdir(parents=True)

class AttrDict(dict):
  def __init__(self):
    dict.__init__(self)
  # Override getattr and setattr so that they return the values of getitem / setitem
  def __setattr__(self, name, value):
    self[name] = value
  def __getattr__(self, name):
    return self[name]

# overwrites args with config parameters
def read_config(args):
  conf = False 
  if os.path.exists(CONFIG_PATH):
    conf = CONFIG_PATH
  if os.path.exists(args.get("CONFIG_PATH")):
    conf = CONFIG_PATH
  if not conf: return args
  POSTS = AttrDict()
  with open(conf) as f:
    for line in f:
      if str(line[0]) == "#" or len(line) <= 1: continue
      try:
        if "\"" in str(line):
          spl = re.split("\"*\"", line)
          key = spl[0] or None
          val = spl[1] or None
        else:
          (key, val) = line.split()
        key = key.strip()
        if str(val)[0] == "\"": val = str(val[1:])
        if str(val)[len(val)-1] == "\"": val = str(val[:len(val)-1])
        # print("{} : {} ".format(key.upper(),val))
        if "_post" in key: 
          setattr(POSTS, key.upper().replace("_post",""), val)
        else: 
          # setattr(args, key.upper(), val)
          # args.set(key.upper(), val)
          args[key.upper()] = val
      except Exception as e:
        print("Warning: Error Parsing Config")
        print(e)
        # pass
  # args.set("POSTS", POSTS)
  args["POSTS"] = POSTS
  return args

##
# Argument Parser
##

parser = argparse.ArgumentParser(prog='OnlySnarf', allow_abbrev=False, epilog="Shnarrf!", 
  description='Post or send messages to OnlyFans.')
# mutually exclusive
# duration & expiration
durationAndExpiration = parser.add_mutually_exclusive_group()
# schedule and date&time
# scheduleAndDate = parser.add_mutually_exclusive_group()
# scheduleAndTime = parser.add_mutually_exclusive_group()
# discount and amount&duration
#
# -action
# the action to be performed
parser.add_argument('-action', type=valid_action, dest='action',
  help='the action to take', choices=ACTIONS, default='post')
##
# -amount
# action: discount
# the amount to discount a user by
parser.add_argument('-amount', type=valid_amount, dest='amount',
  help='the amount (%%) to discount by', default=None)
##
# -backup
# backup uploaded content to "posted" folder
parser.add_argument('-backup', action='store_true', dest='backup',
  help='enables backup processes')
##
# -browser
parser.add_argument('-browser', type=str, default="auto", choices=["auto","google","firefox","auto-remote","remote","remote-chrome","remote-firefox","reconnect"], dest='browser',
  help='the browser to use')
##
# -category
# the category of folder to upload from
parser.add_argument('-category', default=None, dest='category',
  help='the category of content to post or message')
##
# -category-performer
# the category of folder to upload of a performer
parser.add_argument('-category-performer', default=None, dest='performer_category',
  help='the category of content to post or message of a performer')
##
# configurable w/ profile.conf
# OnlySnarf Drive folder list, appends to defaults
parser.add_argument('-categories', dest='categories',
  action='append', help='the categories to list in menu (appends to \'{}\''.format("\'".join(CATEGORIES_DEFAULT)), 
  default=[])
##
# -cookies
# load & save from/to local cookies path
parser.add_argument('-cookies', action='store_true', dest='cookies',
  help='enable loading & saving from/to the local cookies path')
##
# -create-missing 
# creates missing OnlySnarf folders
parser.add_argument('-create-missing', action='store_true', dest='create_missing',
  help='creates missing OnlySnarf folders at target source')
##
# -cron
# determines whether script running is a cronjob
parser.add_argument('-cron', action='store_true', help='toggle cron behavior', dest='cron')
##
# -cron-user
# the user to run OnlySnarf as
parser.add_argument('-cron-user', type=str, dest='cron_user',
  help='the user to run OnlySnarf as', default='root')
##
# -date
# date in MM-DD-YYYY
parser.add_argument('-date', type=valid_date, default=None, dest='date',
  help='schedule date (MM-DD-YYYY)')
##
# -debug
# debugging - skips uploading and deleting unless otherwise forced
parser.add_argument('-debug', action='store_true', dest='debug',
  help='enable debugging')
##
# -debug-force-save
# forces expiration and poll modals to save instead of cancel when debugging
# parser.add_argument('-debug-force-save', action='store_true',
  # help='force expiration and poll to save when debugging')
##
# -debug-delay
# user message delay
parser.add_argument('-debug-delay', action='store_true', dest='debug_delay',
  help='enable a wait between crucial steps for debugging')
##
# -delete-empty
# delete empty content folders
parser.add_argument('-delete-empty', action='store_true', dest='delete_empty',
  help='delete empty content folders')
##
# -delete-google
# delete uploaded content instaed of backing it up
parser.add_argument('-delete-google', action='store_true', dest='delete_google',
  help='delete file instead of backing up')
##
# -destination
# the destination to use when backing up content
parser.add_argument('-destination', dest='destination', default=None, choices=["google",
  # "dropbox",
  "remote","local"],
  help='file backup location. prefers $source if specified')
##
# -discount
# create a format validator for discount where its "[amount]:[duration]"
# parser.add_argument('-discount', metavar='discount', type=valid_discount, default=None,
  # help='discount to apply in format [amount]:[duration]')
##
# download path
parser.add_argument('-download-path', type=str, dest='download_path',
  help='the path to download files to locally', default=DOWNLOAD_PATH)
##
# -duration
# poll duration
parser.add_argument('-duration', type=valid_duration, dest='duration',
  help='the duration in days (99 for \'No Limit\') for a poll', choices=DURATION_ALLOWED, default=None)
##
# -duration-promo
# promotion duration
parser.add_argument('-duration-promo', type=valid_promo_duration, dest='duration_promo',
  help='the duration in days (99 for \'No Limit\') for a promotion', choices=PROMOTION_DURATION_ALLOWED, default=None)
##
# -email
# the OnlyFans email to use for login
parser.add_argument('-email', type=str, default="", dest='email',
  help='the OnlyFans email')
##
# -promotion-expiration
# expiration for a promotion
parser.add_argument('-promotion-expiration', type=int, dest='promotion_expiration',
  help='the promotions expiration in days)', choices=PROMOTION_EXPIRATION_ALLOWED, default=None)
##
# -expiration
# date of post or poll expiration
parser.add_argument('-expiration', type=int, dest='expiration',
  help='the expiration in days (99 for \'No Limit\')', choices=EXPIRATION_ALLOWED, default=None)
##
# -force-backup
# force backup during debugging
parser.add_argument('-force-backup', action='store_true', dest='force_backup',
  help='force backup when debugging')
##
# -force-upload
# ignore upload max wait
parser.add_argument('-force-upload', action='store_true', dest='force_upload',
  help='ignore upload max wait attempts')
##
# -download-max
# maximum number of images to download
parser.add_argument('-download-max', type=int, default=IMAGE_DOWNLOAD_LIMIT, dest="download_limit",
  help='the max number of images to download')
##
# -upload-max
# maximum number of images that can be uploaded
parser.add_argument('-upload-max', type=int, default=IMAGE_UPLOAD_LIMIT, dest='upload_max',
  help='the max number of images to upload')
##
# -keep
# keep the browser window open
parser.add_argument('-keep', action='store_true', dest='keep',
  help='keep the browser window open after script ends')
##
# -keywords
# keywords to # in post
parser.add_argument('-keywords', dest='keywords', action='append', default=[], 
  help="the keywords (#[keyword])")
##
# -limit
# maximum number of subscribers for a promotion
parser.add_argument('-limit', type=valid_limit, default=None, dest='limit', choices=LIMIT_ALLOWED,
  help='the max number of subscribers allowed for a promotion')
##
# -login
# the method to prefer when logging in
parser.add_argument('-login', dest='login', default="auto", choices=["auto","onlyfans","twitter","google"],
  help='the method of login to prefer')
##
# -months
# action: discount
# the number of months to discount for
parser.add_argument('-months', type=valid_month, default=None, dest='months',
  help='the number of months to discount or apply promotion')
##
# -mount-path
# the mounth path for a local directory of OnlyFans config files
parser.add_argument('-mount-path', dest='mount_path',
  help='the local path to OnlySnarf processes')
##
# -bykeyword
# the keyword to search for in folder selection
parser.add_argument('-bykeyword', dest='bykeyword', default=None, 
  help="search for folder by keyword")
##
# -notkeyword
# the keyword to skip in folder selection
parser.add_argument('-notkeyword', dest='notkeyword', default=None,
  help="search for folder not by keyword")
##
# -password
# the password for OnlyFans
parser.add_argument('-password', type=str, dest='password',
  help='the OnlyFans password for login')
##
# -password
# the password for Google
parser.add_argument('-password-google', type=str, dest='password_google',
  help='the Google password for login')
##
# -password
# the password for Twitter
parser.add_argument('-password-twitter', type=str, dest='password_twitter',
  help='the Twitter password for login')
##
# -performers
# list of performers to tag in post
parser.add_argument('-performers', dest='performers', action='append',  default=[],
  help='the performers to list (w/ @[performer]')
# -prefer-local
# prefers local user cache over refreshing first call
parser.add_argument('-prefer-local', action='store_true', dest='prefer_local',
  help='prefer recently cached data')
# -prefer-local-following
# prefers local user cache over refreshing first call for following
parser.add_argument('-prefer-local-following', action='store_true', dest='prefer_local_following',
  help='prefer recently cached data for following')
# -prefer-local-following
# prefers local user cache over refreshing first call for following
# parser.add_argument('-prefer-local-following', action='store_true', dest='prefer_local_following',
#   help='prefer recently cached data for following')
##
# -price
# the price to be set in a message
parser.add_argument('-price', type=valid_price, help='the price', default=0, dest='price')
## 
# -profile-method
parser.add_argument('-profile-method', dest="profile_method", default="syncfrom", choices=["syncto","syncfrom"],
  help='the profile method to use')
##
# -promotion
# the promotion method to use
parser.add_argument('-promotion-method', dest='promotion_method', default="campaign", choices=["campaign","trial","grandfather","user"],
  help='the method of promotion to use')
###
### PATHS ###
# -drive-path
# the folder path within Google Drive for OnlySnarf's root folder
parser.add_argument('-drive-path', dest="drive_path", type=str, 
  help='the folder path within Drive to root OnlySnarf (/OnlySnarf)')
# -config-path
# the path to the config.conf file
parser.add_argument('-config-path', dest="config_path", type=str, 
  help='the path to the config.conf', default=CONFIG_PATH)
# -google-path
# the path to the google_creds.txt
parser.add_argument('-google-creds', dest="google_path", type=str, 
  help='the path to Google credentials', default=GOOGLE_PATH)
# the path to the client_secret.json
parser.add_argument('-client-secret', dest="client_secret", type=str, 
  help='the path to Google secret credentials', default=SECRET_PATH)
# -user-path
# the path to the users.json file
parser.add_argument('-users-path', type=str, dest='users_path',
  help='the path to cache users locally', default=USERS_PATH)
# -profile-path
# the path to the profile.json file
parser.add_argument('-profile-path', type=str, dest='profile_path',
  help='the path to cache profile locally', default=PROFILE_PATH)
##
# -remote-host
# the remote host to connect to
parser.add_argument('-remote-host', type=str, dest='remote_host',
  help='the remote host to connect to', default="127.0.0.1")
##
# -remote-browser-port
# the remote port to connect to
parser.add_argument('-remote-browser-port', type=int, dest='remote_browser_port',
  help='the remote port to connect to for remote browser', default=4444)
##
# -remote-port
# the remote port to connect to
parser.add_argument('-remote-port', type=int, dest='remote_port',
  help='the remote port to connect to', default=22)
##
# -remote-username
# the remote username to use
parser.add_argument('-remote-username', type=str, dest='remote_username',
  help='the remote username to use', default=None)
##
# -remote-password
# the remote password to use
parser.add_argument('-remote-password', type=int, dest='remote_password',
  help='the remote password to use', default=None)
###
##
# -question
# poll questions
parser.add_argument('-question', dest='questions', action='append', default=[],
  help='the questions to ask')
###
# the maximum number of recent users
parser.add_argument('-recent-users-count', default=3, dest='recent_users_count',
  type=int, help='the number of users to consider recent')
##
# enables file reduction
parser.add_argument('-reduce', action='store_true', dest='reduce',
  help='enable reducing files under 50 MB')
##
# enables file repair (buggy)
parser.add_argument('-repair', action='store_true', dest='repair',
  help='enable repairing videos as appropriate')
##
# can be set in profile.conf
# root Google drive folder
parser.add_argument('-drive-root', type=str, default='OnlySnarf', dest='drive_root',
  help='the Google Drive root folder name')
##
# -save-users
# saves OnlyFans users upon exit
parser.add_argument('-save-users', action='store_true', dest='save_users',
  help='enable saving users locally on exit')
##
# -schedule
# 
parser.add_argument('-schedule', type=valid_schedule, default=None, dest='schedule',
  help='the schedule (MM-DD-YYYY:HH:MM)')
##
# -session-id
parser.add_argument('-session-id', default=None, dest='session_id',
  help='the session id to use')
# -session-url
parser.add_argument('-session-url', default=None, dest='session_url',
  help='the session url to use')
##
# -skip-download
parser.add_argument('-skip-download', action='store_true', dest='skip_download',
  help='skip file downloads')
##
# -skip-upload
# skips file upload
parser.add_argument('-skip-upload', action='store_true', dest='skip_upload',
  help='skip file uploads')
##
# list of users to skip
parser.add_argument('-skip-users', dest='skipped_users', 
  action='append', help='the users to skip or ignore ')
##
# -show
# shows window
parser.add_argument('-show','-show-window', dest='show', action='store_true', 
  help='enable displaying the browser window')
##
# -sort
# the sort method to use when selecting content unprompted
parser.add_argument('-sort', dest='sort', default="random", choices=["ordered","random"],
  help='the sort method to use when selecting content unprompted')
##
# -source
# the source to use when searching for content
parser.add_argument('-source', dest='source', default=None, choices=["google",
  # "dropbox",
  "remote","local"],
  help='file host location')
##
# -tags
# @[tag]
parser.add_argument('-tags', dest='tags', action='append', default=[],
  help='the tags (@[tag])')
##
# -text
# text for message or upload
parser.add_argument('-text', default=None, dest='text',
  help='the text to type')
##
# -time
# time in HH:MM
parser.add_argument('-time', type=valid_time, default=None, dest='time',
  help='the time (HH:MM)')
##
# -title
# the title of a file to search for
parser.add_argument('-title', default=None, dest='title',
  help='the title of the file to search for')
##
# -thumbnail
parser.add_argument('-thumbnail', action='store_true', dest='thumbnail',
  help='fix thumbnails when necessary')
##
# -tweet
# enabled tweeting
parser.add_argument('-tweeting', action='store_true', dest='tweeting',
  help='enable tweeting when posting')
##
# -upload-max
# the max number of 10 minute intervals to upload for
parser.add_argument('-upload-max-duration', dest='upload_max_duration', default=UPLOAD_MAX_DURATION,
  type=int, help='the number of 10 minute intervals to wait while uploading a file')
##
# -user
# the user to target
parser.add_argument('-user', type=str,  default=None, dest='user',
  help='the user to message')
##
# -users
# the users to target
parser.add_argument('-users', dest='users', action='append', default=[],
  help='the users to message')
##
# -users-favorite
# list of favorited users
parser.add_argument('-users-favorite', default=[],
  dest='users_favorite', action='append', help='supplied list of favorite users')
##
# -username
# the OnlyFans username to use
parser.add_argument('-username', type=str, default="", dest='username',
  help='the OnlyFans username')
##
# -username-google
# the Google username to use
parser.add_argument('-username-google', type=str, default="", dest='username_google',
  help='the Google username for login')
##
# -username-twitter
# the Twitter username to use
parser.add_argument('-username-twitter', type=str, default="", dest='username_twitter',
  help='the Twitter username for login')
##
# -users-read
# the number of users read when checking messages
parser.add_argument('-users-read', type=int, dest='users_read',
  help='the number of users to read when checking messages', default=10)
##
# -verbose
# v, vv, vvv
parser.add_argument('-v', '-verbose', dest="verbose", action='count', default=0, 
  help="verbosity level (max 3)")

## Promotion Methods
# -promotion-user
parser.add_argument('-promotion-user', dest="promotion_user", action='store_true', 
  help="uses user method when combined with action=promotion")



##
# Positional
##

##
# input
parser.add_argument('input', default=[], nargs=argparse.REMAINDER, 
  type=valid_path, help='file input to post or message')

##
import pkg_resources
parser.version = str(pkg_resources.get_distribution("onlysnarf").version)
parser.add_argument('-version', action='version')

############################################################################################

try:
  args = vars(parser.parse_args())
except Exception as e:
  print("Error: Incorrect arg format")
  import sys
  sys.exit(1)

if args["source"] and not args["destination"]:
  args["destination"] = args["source"]
if not args["source"]: args["source"] = "local"
if not args["destination"]: args["destination"] = "local"
# if args["debug"]: ACTIONS.append("promotion")
# if args["debug"]: ACTIONS.append("test")
if not args["category"]: args["source"] = None

############################################################################################

# print(args)
CONFIG = {}
for key in args:
  CONFIG[key.upper()] = args.get(key)
CONFIG = read_config(CONFIG)

#############
# Debugging #
# import sys
# print(CONFIG)
# sys.exit(0)


# from .src.classes import Discount, Poll, Promotion, Schedule
# from .src.driver import Driver
# from .src.settings import Settings
# from .src.user import User
# from .src.profile import Profile
# from .src.snarf import Snarf
# from .bin import config as Config
# from .bin import menu as Menu


#!/usr/bin/python3
# main OnlySnarf class

import random
import os
import shutil
import datetime
import json
import sys
import pathlib
import time
##
from .settings import Settings

#####################
##### OnlySnarf #####
#####################

class Snarf:
    """OnlySnarf main class and runtime parser"""

    def __init__(self):
        """Snarf object"""

        pass

    @staticmethod
    def discount(discount=None):
        """
        Applies the provided discount or creates one from args / prompts.

        Parameters
        ----------
        discount : classes.Discount
            A discount consisting of amount, months, and / or username. Prompts for missing.

        """
        from .classes import Discount
        if not discount: discount = Discount()
        try: discount.apply()
        except Exception as e: Settings.dev_print(e)

    @staticmethod
    def message(message=None):
        """
        Sends the provided message or creates one from args / prompts.

        Parameters
        ----------
        message : classes.Message
            A message consisting of text, recipient(s), and possibly also files, keywords, tags, 
                performers, and/or price 
        
        """
        from .classes import Message
        from .user import User
        if not message: message = Message()
        try:
            message.get_message()
            if Settings.is_prompt():
                if not Settings.prompt("Send"): return
            if message.get_files() != "unset" and len(message.get_files()) == 0 and not message.get_text():
                Settings.err_print("Missing Files and Text")
                return
            successful = False
            try: 
                # for user in self.get_recipients():
                for user in message.users:
                    # if isinstance(user, str) and str(user) == "post": successful_ = Driver.post(self)
                    # print("Messaging: {}".format(user.username))
                    if isinstance(user, User): successful = User.message_user(username=user.username, message=message)
                    else: successful = User.message_user(username=user, message=message)
            except Exception as e:
                Settings.dev_print(e)
                successful = False
            if successful: message.cleanup_files()
        except Exception as e: Settings.dev_print(e)
                
    @staticmethod
    def post(message=None):
        """
        Posts the provided text or from args / prompts.

        Parameters
        ----------
        message : classes.Post
            A post consisting of text and possibly also files, keywords, tags, performers,
                expiration, poll, and/or schedule
        
        """
        from .classes import Message
        if not message: message = Message()
        try:
            message.get_post()
            if Settings.is_prompt():
                if not Settings.prompt("Post"): return
            if message.get_files() != "unset" and len(message.get_files()) == 0 and not message.get_text():
                Settings.err_print("Missing Files and Text")
                return
            successful = False
            try:
                from .driver import Driver
                successful = Driver.get_driver().post(message=message)
            except Exception as e:
                Settings.dev_print(e)
                successful = False
            if successful: message.cleanup_files()
        except Exception as e: Settings.dev_print(e)

    @staticmethod
    def profile(profile=None):
        """
        Runs the profile method specified at runtime.

        backup - downloads all content and saves settings

        syncFrom - reads all profile settings and saves locally

        syncTo - updates profile settings with provided profile

        Extended description of function.

        Parameters
        ----------
        profile : profile.Profile
            Class representation of Onlyfans profile settings

        """
        from .profile import Profile
        if not profile: profile = Profile()
        try: 
            # get profile method
            method = Settings.get_profile_method()
            if method == "backup":
                profile.backup_content()
            elif method == "syncfrom":
                Profile.sync_from_profile()
            elif method == "syncto":
                Profile.sync_to_profile()
            else: Settings.err_print("Missing Profile Method")
        except Exception as e: Settings.dev_print(e)
        
    @staticmethod
    def promotion():
        """
        Runs the promotion method specified at runtime.

        campain - creates discount campaign

        trial - creates free trial

        user - applies directly to user

        grandfather - applies discounted price to existing users and adds them all to list

        """
        from .classes import Promotion
        try: 
            # get promotion method
            method = Settings.get_promotion_method()
            if method == "campaign":
                Promotion.create_campaign()
            elif method == "trial":
                Promotion.create_trial_link()
            elif method == "user":
                Promotion.apply_to_user()
            elif method == "grandfather":
                Promotion.grandfathered()
            else: Settings.err_print("Missing Promotion Method")
        except Exception as e: Settings.dev_print(e)

    # developer testing
    @staticmethod
    def test():
        from .user import User
        print('1/3 : Testing')
        print('TESTING: Settings - Get')
        profile = Profile.sync_from_profile()
        print('TESTING: Settings - Set')
        Profile.sync_to_profile(profile=profile)
        return True

################################################################################################################################################

def exit_handler():
    """Exit cleanly"""

    from .driver import Driver
    Driver.exit_all()
    print("Shnarrf?")
    sys.exit(0)

import atexit
atexit.register(exit_handler)

def main():
    try:
        from .file import File
        # File.remove_local()
        Settings.set_prompt(False)
        Settings.set_confirm(False)
        action = Settings.get_action()
        print("Running - {}".format(action))
        ## Actions
        success = False
        if str(action) == "test":
            success = Snarf.test()
        elif str(action) == "bot":
            from .bot import Bot
            bot = Bot()
            success = bot.run()
        elif str(action) == "post":
            success = Snarf.post()
        elif str(action) == "message":
            success = Snarf.message()
        elif str(action) == "discount":
            success = Snarf.discount()
        elif str(action) == "promotion":
            success = Snarf.promotion()
        elif str(action) == "profile":
            success = Snarf.profile()
        else:
            Settings.warn_print("Missing Method")
    except Exception as e:
        Settings.dev_print(e)
        print("Shnarf!")
    finally:
        exit_handler()

################################################################################################################################################

if __name__ == "__main__":
    main()


import os, shutil, random, sys
from .ffmpeg import ffmpeg
from . import google as Google
from .settings import Settings
import PyInquirer
from . import remote as Remote
from PIL import Image
from os import walk

ONE_GIGABYTE = 1000000000
ONE_MEGABYTE = 1000000
FIFTY_MEGABYTES = 50000000
ONE_HUNDRED_KILOBYTES = 100000

MIMETYPES_IMAGES = "(mimeType contains 'image/jpeg' or mimeType contains 'image/jpg' or mimeType contains 'image/png')"
MIMETYPES_VIDEOS = "(mimeType contains 'video/mp4' or mimeType contains 'video/quicktime' or mimeType contains 'video/x-ms-wmv' or mimeType contains 'video/x-flv')"
MIMETYPES_ALL = "(mimeType contains 'image/jpeg' or mimeType contains 'image/jpg' or mimeType contains 'image/png' or mimeType contains 'video/mp4' or mimeType contains 'video/quicktime')"

MIMETYPES_IMAGES_LIST = ["image/jpeg","image/jpg","image/png"]
MIMETYPES_VIDEOS_LIST = ["video/mp4","video/quicktime","video/x-ms-wmv","video/x-flv"]
MIMETYPES_ALL_LIST = []
MIMETYPES_ALL_LIST.extend(MIMETYPES_IMAGES_LIST)
MIMETYPES_ALL_LIST.extend(MIMETYPES_VIDEOS_LIST)

def print_same_line(text):
    sys.stdout.write('\r')
    sys.stdout.flush()
    sys.stdout.write(text)
    sys.stdout.flush()

###############################################################

class File():
    """File class"""

    FILES = None

    def __init__(self):
        """File object represents local image/video file"""

        # the path to the file locally
        self.path = None
        # the file extension
        self.ext = None
        # image|video
        self.type = None
        ##
        # file title reference
        self.title = None
        # [image, gallery, video, performer]
        self.category = None
        # file size
        self.size = None

    ######################################################################################

    def backup(self):
        """Backup file to appropriate destination source"""

        if not File.backup_text(self.get_title()): return
        if Settings.get_destination() == "remote":
            Remote.upload_file(self)
        elif Settings.get_destination() == "google":
            Google.upload_file(file=self)
        else:
            # move file to local backup location
            backupPath = os.path.join(Settings.get_local_path(), "posted")
            backupPath = os.path.join(backupPath, self.category, self.get_title())
            shutil.move(self.get_path(), backupPath)

    @staticmethod
    def backup_text(title):
        """
        Print applicable backup text

        Returns
        -------
        bool
            Whether or not the file should be backed up

        """

        if Settings.is_skip_download():
            Settings.maybe_print("Warning: Skipping Backup, skipped download")
            return False
        if Settings.is_force_backup():
            Settings.maybe_print("Backing Up (forced): {}".format(title))
        elif not Settings.is_backup():
            Settings.maybe_print('Skipping Backup (disabled): {}'.format(title))
            return False
        elif Settings.is_debug():
            Settings.maybe_print("Skipping Backup (debug): {}".format(title))
            return False
        else:
            Settings.maybe_print('Backing Up: {}'.format(title))
        return True

    @staticmethod
    def backup_files(files=[]):
        """
        Backup files provided to appropriate destinations

        Returns
        -------
        bool
            Whether or not the files were backed up successfully

        """

        if not File.backup_text(self.get_title()): return
        if Settings.get_destination() == "remote":
            Remote.upload_files(files)
        elif Settings.get_destination() == "google":
            Google.upload_files(files)
        else:
            for file in files:
                file.backup()
        return True

    def check_size(self):
        """
        Check file size.

        Returns
        -------
        bool
            Whether or not the file exists by checking size

        """

        if not self.size:
            if not os.path.exists(self.get_path()): return False
            size = os.path.getsize(self.get_path())
        else: size = self.size
        Settings.maybe_print("File Size: {}kb - {}mb".format(size/1000, size/1000000))
        global ONE_MEGABYTE
        if size <= ONE_MEGABYTE:
            Settings.warn_print("Small File Size")
        global ONE_HUNDRED_KILOBYTES
        if size <= ONE_HUNDRED_KILOBYTES:
            Settings.warn_print("Tiny File Size")
        self.size = size
        if size == 0:
            Settings.err_print("Empty File Size")
            return False
        return True

    def delete(self):
        """Delete file"""

        if not File.delete_text(self.get_title()): return
        try: 
            os.remove(self.get_path())
            print('File Deleted: {}'.format(self.get_title()))
        except Exception as e: Settings.dev_print(e)

    @staticmethod
    def delete_text(title):
        """
        Print applicable deletion text
        
        Returns
        -------
        bool
            Whether or not the file should be deleted

        """

        if Settings.is_skip_download():
            Settings.maybe_print("Warning: Skipping Delete, skipped download")
            return False
        if not Settings.is_delete():
            Settings.maybe_print('Skipping Delete (disabled): {}'.format(title))
            return False
        elif Settings.is_debug():
            Settings.maybe_print("Skipping Delete (debug): {}".format(title))
            return False
        else:
            Settings.maybe_print('Deleting: {}'.format(title))
        return True

    @staticmethod
    def download_text(title):
        """
        Print applicable download text.
        
        Returns
        -------
        bool
            Whether or not the file should be downloaded

        """

        if Settings.is_skip_download():
            print("Skipping Download (debug)")
            return False
        return True

    ##############################

    def get_ext(self):
        """Get the file's extension"""

        if self.ext: return self.ext
        self.get_title()

    def get_path(self):
        """
        Get the file's path
        
        Returns
        -------
        str
            The file path

        """

        if not self.path:
            Settings.maybe_print("Error: Missing File Path")
            return  ""
        return self.path

    def get_title(self):
        """
        Get the file's title from it's filename
        
        Returns
        -------
        str
            The file's title or filename without extension

        """

        if self.title: return self.title
        path = self.get_path()
        if str(path) == "": 
            Settings.maybe_print("Error: Missing File Title")
            return ""
        title, ext = os.path.splitext(path)
        self.ext = ext
        self.title = "{}{}".format(os.path.basename(title), ext)
        return self.title

    @staticmethod
    def get_tmp():
        """Creates / gets the default temporary download directory"""

        # tmp = os.getcwd()
        # if Settings.get_download_path() != "":
        #     tmp = os.path.join(Settings.get_download_path(), "tmp")
        # else:
        #     tmp = os.path.join(tmp, "tmp")
        # if not os.path.exists(str(tmp)):
        #     os.mkdir(str(tmp))
        # return tmp
        download_path = Settings.get_download_path()
        if not os.path.exists(str(download_path)):
            os.mkdir(str(download_path))
        return download_path

    def get_type(self):
        """
        Gets the file's type as an inner class of either Image or Video
        
        Returns
        -------
        Image|Video
            The file's type as an image or video class

        """

        if self.type: return self.type
        if str(self.get_ext()) in str(MIMETYPES_VIDEOS_LIST):
            self.type = Video()
        elif str(self.get_ext()) in str(MIMETYPES_IMAGES_LIST):
            self.type = Image()
        else: print("Warning: Unable to Parse File Type")
        return self.type

    def prepare(self):
        """
        Prepares the file for uploading.

        Runs the apppropriate file type method and downloads the file locally if necessary.

        Returns
        -------
        bool
            Whether or not the file is prepared

        """

        # Settings.maybe_print("preparing: {}".format(self.get_title()))
        self.get_type().prepare()
        if not self.check_size():
            return False
        return True

    @staticmethod
    def remove_local():
        """
        Delete all local files.

        """

        try:
            # if str(Settings.SKIP_DELETE) == "True":
                # Settings.maybe_print("Skipping Local Remove")
                # return
            # print('Deleting Local File(s)')
            # delete /tmp
            tmp = File.get_tmp()
            if os.path.exists(tmp):
                shutil.rmtree(tmp)
                print('Local File(s) Removed')
            else:
                print('Local Files Not Found')
        except Exception as e:
            Settings.dev_print(e)

    @staticmethod
    def get_files():
        """
        Get files from the runtime category folder.

        Returns
        -------
        list
            The files retrieved

        """

        if File.FILES: return File.FILES
        category = Settings.get_category()
        if not category: category = Settings.select_category()
        if not category: Settings.dev_print("Warning: Missing Category")
        files = File.get_files_by_category(category)
        if Settings.get_title() and str(files) != "unset":
            for file in files:
                if str(Settings.get_title()) == str(file.get_title()):
                    files = [file]
                    break
        File.FILES = files
        return files

    @staticmethod
    def get_files_by_folder(path):
        """
        Get local files from the local folder path.

        Parameters
        ----------
        path : str
            Path to folder to get files of

        Returns
        -------
        list
            The files at the path

        """

        f = []
        for (dirpath, dirnames, filenames) in walk(path):
            f.extend(filenames)
            break
        return f

    def get_folder_by_name(category, parent=None):
        """
        Get local folder by category and parent.

        Parameters
        ----------
        category : str
            The category or folder name to get
        parent : file.Folder
            The local folder's parent to search within

        Returns
        -------
        str
            The folder path of the found folder

        """

        if not parent:
            parent = Settings.get_local_path()
        Settings.maybe_print("parent: {}".format(parent))
        f = []
        for (dirpath, dirnames, filenames) in walk(parent):
            Settings.dev_print("dirpath: {}".format(dirpath))
            for dir_ in dirnames:
                Settings.dev_print("dir: {} = {} :category".format(dir_, category))
                if str(dir_) == str(category):
                    return os.path.join(parent, dir_)
            break
        return None













































































































































































    @staticmethod
    def get_folders_of_folder_by_keywords(categoryFolder):
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        Settings.dev_print("getting keywords of folder: {}".format(categoryFolder))
        if categoryFolder == None: return []
        folders = File.get_folders_of_folder(categoryFolder)
        folders_ = []
        for folder in folders:
            if Settings.get_drive_keyword() and str(folder.get_title()) != str(Settings.get_drive_keyword()):
                Settings.dev_print('{} -> not keyword'.format(folder.get_title()))
                continue
            elif Settings.get_drive_ignore() and str(folder.get_title()) == str(Settings.get_drive_ignore()):
                Settings.dev_print('{} -> by not keyword'.format(folder.get_title()))
                continue
            elif str(folder.get_title()) == str(Settings.get_drive_keyword):
                Settings.dev_print('{} -> by keyword'.format(folder.get_title()))
            else:
                Settings.dev_print("{}".format(folder.get_title()))
            folders_.append(folder)
        return folders_

    @staticmethod
    def get_random_file():
        """Get random file from all files"""

        return random.choice(File.get_files())

    @staticmethod
    def get_images_of_folder(folder):
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        Settings.dev_print("getting images of folder: {}".format(folder.get_title()))
        if not folder: return []
        imgs = []
        files = []
        valid_images = [".jpg",".gif",".png",".tga",".jpeg"]
        for f in os.listdir(folder.get_path()):
            ext = os.path.splitext(f)[1]
            if ext.lower() not in valid_images:
                continue
            file = File()
            setattr(file, "path", os.path.join(folder.get_path(),f))
            files.append(file)
            Settings.maybe_print("image path: {}".format(os.path.join(folder.get_path(),f)))
        return files

    @staticmethod
    def get_videos_of_folder(folder):
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        Settings.dev_print("getting videos of folder: {}".format(folder.get_title()))
        if not folder: return []
        videos = []
        files = []
        valid_videos = [".mp4",".mov"]
        for f in os.listdir(folder.get_path()):
            ext = os.path.splitext(f)[1]
            if ext.lower() not in valid_videos:
                continue
            file = File()
            setattr(file, "path", os.path.join(folder.get_path(),f))
            files.append(file)
            Settings.maybe_print("video path: {}".format(os.path.join(folder.get_path(),f)))
        return files

    @staticmethod
    def get_folders_of_folder(folderPath):
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        # os.walk(directory)
        # will yield a tuple for each subdirectory. Ths first entry in the 3-tuple is a directory name, so
        # [x[0] for x in os.walk(directory)]
        # should give you all of the subdirectories, recursively.
        # Note that the second entry in the tuple is the list of child directories of the entry in the first position, so you could use this instead, but it's not likely to save you much.
        # However, you could use it just to give you the immediate child directories:
        Settings.maybe_print("local walk: {}".format(folderPath))
        folders = []
        # print(os.walk(folderPath))
        for folder in next(os.walk(folderPath))[1]:
            Settings.maybe_print("folder: {}".format(folder))
            fol = Folder()
            setattr(fol, "path", os.path.join(folderPath, folder))
            folders.append(fol)
        return folders

    @staticmethod
    def get_files_by_category(cat, performer=None):
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        Settings.maybe_print("loading local files...")
        files = []
        ##
        def parse_categories(category, categoryFolder=None):
            """
            Summary line.

            Extended description of function.

            Parameters
            ----------
            arg1 : int
                Description of arg1
            arg2 : str
                Description of arg2

            Returns
            -------
            int
                Description of return value

            """

            files = []
            # return File.get_files_by_category(cat)
            if "image" in str(category):
                categoryFolder = File.get_folder_by_name(category, parent=categoryFolder)
                for folder in File.get_folders_of_folder_by_keywords(categoryFolder):
                    if not folder: continue
                    for image in File.get_images_of_folder(folder):
                        file = File()
                        setattr(file, "path", imageget_path())
                        setattr(file, "category", folder.get_title())
                        files.append(file)
            elif "video" in str(category):
                categoryFolder = File.get_folder_by_name(category, parent=categoryFolder)
                for folder in File.get_folders_of_folder_by_keywords(categoryFolder):
                    if not folder: continue
                    videos = File.get_videos_of_folder(folder)
                    # if len(videos) > 0:
                        # files.append(folder)
                    for video in videos:
                        file = File()
                        setattr(file, "path", video.get_path())
                        setattr(file, "category", folder.get_title())
                        files.append(file)
            elif "performer" in str(category):
                categoryFolder = File.get_folder_by_name(category, parent=categoryFolder)
                for performer_ in File.get_folders_of_folder_by_keywords(categoryFolder):
                    # for performer in File.get_folders_of_folder(folder):
                    if not performer_: continue
                    p = Folder()
                    setattr(p, "path", performer_.get_path())
                    setattr(p, "category", categoryFolder.get_title())
                    files.append(p)
            # elif "galler" in str(category):
            else:
                categoryFolder = File.get_folder_by_name(category, parent=categoryFolder)
                for folder in File.get_folders_of_folder_by_keywords(categoryFolder):
                    if not folder: continue
                    galleries = File.get_folders_of_folder(folder)
                    if len(galleries) > 0:
                        files.append(folder)
                    for gallery in galleries:
                        file = Folder()
                        setattr(file, "path", galleryget_path())
                        setattr(file, "category", folder.get_title())
                        files.append(file)
            return files
        ##
        if performer:
            categoryFolder = File.get_folder_by_name("performers")
            for performerFolder in File.get_folders_of_folder_by_keywords(categoryFolder):
                if str(performer) == str(performerFolder.get_title()):
                    return parse_categories(cat, categoryFolder=performerFolder)
        return parse_categories(cat)

    @staticmethod
    def select_file(category, performer=None):
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        files = File.get_files_by_category(category, performer=performer)
        files_ = []
        for file in files:
            if isinstance(file, str):
                files_.append(PyInquirer.Separator())
                continue
            file.category = category
            file_ = {
                "name": file.get_title(),
                "value": file,
            }
            files_.append(file_)
        if len(files_) == 0:
            print("Missing Files")
            return
        files_.append({
            "name": 'Back',
            "value": None,
        })
        question = {
            'type': 'list',
            'name': 'file',
            'message': 'File Path:',
            'choices': files_,
            # 'filter': lambda file: file.lower()
        }
        answer = PyInquirer.prompt(question)
        if not answer: return File.select_files()
        file = answer["file"]
        if not Settings.confirm(file.get_path()): return None
        return file

    @staticmethod
    def select_files():
        """
        Summary line.

        Extended description of function.

        Parameters
        ----------
        arg1 : int
            Description of arg1
        arg2 : str
            Description of arg2

        Returns
        -------
        int
            Description of return value

        """

        if not Settings.is_prompt(): return [File.get_random_file()]
        category = Settings.select_category()
        if not category: return File.select_file_upload_method()
        # if not Settings.confirm(category): return File.select_files()
        print("Select Files or a Folder")
        files = []
        while True:
            file = File.select_file(category)
            if not file: break
            ##
            if "performer" in str(category):
                cat = Settings.select_category([cat for cat in Settings.get_categories() if "performer" not in cat])
                performerName = file.get_title()
                file = File.select_file(cat, performer=performerName)
                if not file: break
                setattr(file, "performer", performerName)
                files.append(file)
                if "galler" in str(cat) or "video" in str(cat): break
            ##
            files.append(file)
            if "galler" in str(category) or "video" in str(category): break
        if str(files[0]) == "unset": return files
        if not Settings.confirm([file.get_title() for file in files]): return File.select_files()
        return files

    @staticmethod
    def select_file_upload_method():
        """
        Menu to select the method to upload a file.

        Returns
        -------
        list
            The appropriately selected files

        """

        if not Settings.prompt("upload files"): 
            return "unset"
        print("Select an upload source")
        sources = Settings.get_source_options()
        question = {
            'type': 'list',
            'name': 'upload',
            'message': 'Upload:',
            'choices': [src.title() for src in sources]
        }
        upload = PyInquirer.prompt(question)["upload"]


        # everything after this part should be in another function
        # this should just return the string of the upload source


        if str(upload) == "Local":
            return File.select_files()
        elif str(upload) == "Google":
            return Google_File.select_files()
        # elif str(upload) == "Dropbox":
            # return Dropbox.select_files()
        elif str(upload) == "Remote":
            return Remote.select_files()
        return File.select_files()















    def upload(self):
        """
        Process ran by a file after it has been uploaded.

        Ensures the file has been backed up and then deleted locally.
        
        Returns
        -------
        bool
            Whether or not the file was properly handled after its upload

        """
        if not self.prepare():
            print("Error: Unable to Upload File - {}".format(self.get_title()))
            return False
        self.backup()
        self.delete()
        return True

##


































class Remote_File(File):
    def __init__(self):
        File.__init__(self)

    def backup(self):
        if not File.backup_text(self.get_title()): return
        if Settings.get_destination() == "remote":
            Remote.backup_file(self)
        elif Settings.get_destination() == "google":
            Google.upload_file(file=self)
        # elif Settings.get_destination() == "dropbox":
        #     Dropbox.upload_file(file=self)
        else:
            file = self.download()
            file.backup()
            self.delete()

    def delete(self):
        if not File.delete_text(self.get_title()): return
        try: 
            Remote.delete_file(self)
            print('File Deleted: {}'.format(self.get_title()))
        except Exception as e: Settings.dev_print(e)

    def download(self):
        if not File.download_text(self.get_title()): return False
        file = Remote.download_file(self)
        if not file: return False
        ### Finish ###
        if not file.check_size():
            print("Error: Missing Downloaded File")
            return False
        print("Downloaded: {}".format(file.get_title()))
        return file

##

class Folder(File):
    def __init__(self):
        File.__init__(self)
        self.files = None

    def backup(self):
        if File.backup_text(self.get_title()): return
        # Google.upload_gallery(files=self.files)

    def check_size(self):
        for file in self.get_files():
            exists = file.check_size()
            if not exists: return False
        return True

    # def combine(self):
    #     if len(self.files) == 0: return
    #     Settings.dev_print("combining files: {}".format(len(self.files)))
    #     Settings.dev_print("combine path: {}".format(combinedPath))
    #     combinedPath = os.path.join(File.get_tmp(), "{}-combined".format(self.title))
    #     for file in files:
    #         shutil.move(file.get_path(), combinedPath)
    #         file.path = "{}/{}".format(combinedPath, self.title)
    #     self.combined = ffmpeg.combine(combinedPath)

    ##############################


    def download(self):
        print("Downloading Folder: {}".format(self.get_title()))
        if len(self.files) == 0:
            file_list = Google.get_files_by_folder_id(self.get_id())
            self.files = []
            for file in file_list:
                file_ = Google_File()
                setattr(file_, "id", file["id"])
                setattr(file_, "file", file)
                self.files.append(file_)
        folder_size = len(self.files)
        Settings.maybe_print('Folder size: {}'.format(folder_size))
        Settings.maybe_print('Upload limit: {}'.format(Settings.get_upload_max()))
        if int(folder_size) == 0:
            print('Error: Empty Folder')
            return False
        file_list = self.files
        random.shuffle(file_list)
        file_list = file_list[:int(Settings.get_upload_max())]
        ## video preference
        videos = []
        for file in file_list:
            if str(file.get_mimetype()) in MIMETYPES_VIDEOS_LIST:
                videos.append(file)
        if len(videos) > 0: file_list = [random.choice(videos)]
        ##
        i = 1
        for file in sorted(file_list, key = lambda x: x.get_title()):
            # print_same_line("Downloading: {} ({}/{})".format(file.get_title(), i, folder_size))
            print("Downloading: {} ({}/{})".format(file.get_title(), i, folder_size))
            file.download()
            i+=1
        print()
        print("Downloaded Folder: {}".format(self.get_title()))

    def get_files(self):
        if not self.files and self.path:
            self.files = []
            files = File.get_files_by_folder(self.get_path())
            for file in files:
                file_ = File()
                setattr(file_, "path", os.path.join(self.get_path(), file))
                self.files.append(file_)
                Settings.maybe_print("local file found: {}".format(file_.get_title()))
        if Settings.get_title():
            for file in self.files:
                if str(Settings.get_title()) == str(file.get_title()):
                    self.files = [file]
                    break
        return self.files

    def get_title(self):
        if self.title: return self.title
        path = self.get_path()
        if str(path) == "": 
            Settings.maybe_print("Error: Missing File Title")
            return ""
        title = os.path.basename(path)
        self.title = title
        return self.title

    def prepare():
        prepared = False
        for file in self.get_files():
            prepared_ = file.prepare()
            if prepared_: prepared = prepared_
        return prepared


###################################################################################

class Google_File(File):

    def __init__(self):
        File.__init__(self)
        self.id = None
        self.title = None
        self.file = None
        self.parent = None
        self.mimeType = None

    def backup(self):
        if not File.backup_text(self.get_title()): return
        Google.backup_file(self)

    def delete(self):
        if not File.delete_text(self.get_title()): return
        Google.delete(self)

    @staticmethod
    def download_files(files=[]):
        Settings.maybe_print('Download limit: '+str(Settings.get_image_download_limit()))
        random.shuffle(files)
        files = files[:int(Settings.get_image_download_limit())]
        print('Downloading Files: {}'.format(len(files)))
        i = 1
        for file in sorted(files, key = lambda x: x['title']):
            print('Downloading: {}/{}'.format(i, Settings.get_image_download_limit()))
            file.download()
        print("Downloaded: {}".format(len(files)))

    # Download File
    def download(self):
        if not File.download_text(self.get_title()): return False
        successful = Google.download_file(self)
        if not successful: return False
        ### Finish ###
        if not self.check_size():
            print("Error: Missing Downloaded File")
            return False
        print("Downloaded: {}".format(self.get_title()))
        return True

    def get_ext(self):
        if self.ext: return self.ext
        title, ext = os.path.splitext(self.get_file()["title"])
        if str(ext) == "":
            ext = self.get_file()["mimeType"]
            mime, ext = str(ext).split("/")
            ext = "."+str(ext)
        self.ext = ext
        self.title = title
        return self.ext

    def get_id(self):
        if self.id: return self.id
        if self.file: self.id = self.file["id"]
        return self.id

    def get_file(self):
        if self.file: return self.file
        self.file = Google.get_file(self.get_id())
        # if not self.check_size(): self.download()
        return self.file

    @staticmethod
    def get_files():
        if File.FILES: return File.FILES
        category = Settings.get_category()
        if not category: category = Settings.select_category()
        if not category: Settings.dev_print("Warning: Missing Category")
        files = Google_File.get_files_by_category(category)
        if Settings.get_title():
            for file in files:
                if str(Settings.get_title()) == str(file.get_title()):
                    files = [file]
                    break
        File.FILES = files
        return files

    @staticmethod
    def get_files_by_category(cat, performer=None):
        Settings.maybe_print("Loading Google Files...")
        files = []
        ##
        def parse_categories(category, categoryFolder=None):
            files = []
            # return Google_File.get_files_by_category(cat)
            if "image" in str(category):
                categoryFolder = Google.get_folder_by_name(category, parent=categoryFolder)
                for folder in Google.get_folders_of_folder_by_keywords(categoryFolder):
                    for image in Google.get_images_of_folder(folder):
                        file = Google_File()
                        setattr(file, "file", image)
                        setattr(file, "parent", folder)
                        if performer: setattr(file, "performer", performer)
                        files.append(file)
            elif "video" in str(category):
                categoryFolder = Google.get_folder_by_name(category, parent=categoryFolder)
                for folder in Google.get_folders_of_folder_by_keywords(categoryFolder):
                    videos = Google.get_videos_of_folder(folder)
                    if len(videos) > 0:
                        files.append(folder["title"])
                    for video in videos:
                        file = Google_File()
                        setattr(file, "file", video)
                        setattr(file, "parent", folder)
                        if performer: setattr(file, "performer", performer)
                        files.append(file)
            elif "galler" in str(category):
                categoryFolder = Google.get_folder_by_name(category, parent=categoryFolder)
                for folder in Google.get_folders_of_folder_by_keywords(categoryFolder):
                    galleries = Google.get_folders_of_folder(folder)
                    if len(galleries) > 0:
                        files.append(folder["title"])
                    for gallery in galleries:
                        file = Google_Folder()
                        setattr(file, "file", gallery)
                        setattr(file, "parent", folder)
                        if performer: setattr(file, "performer", performer)
                        files.append(file)
            elif "performer" in str(category):
                categoryFolder = Google.get_folder_by_name(category, parent=categoryFolder)
                for performer_ in Google.get_folders_of_folder_by_keywords(categoryFolder):
                    # for performer in Google.get_folders_of_folder(folder):
                    p = Google_Folder()
                    setattr(p, "file", performer_)
                    setattr(p, "parent", categoryFolder)
                    files.append(p)
            return files
        ##
        if str(cat) == "performers":
            categoryFolder = Google.get_folder_by_name("performers")
            performerFolders = Google.get_folders_of_folder_by_keywords(categoryFolder)
            for performerFolder in performerFolders:
                if performer and str(performer) == str(performerFolder['title']):
                    return parse_categories(cat, categoryFolder=performerFolder)

            if Settings.get_sort_method() == "ordered":
                sortedPerformers = performerFolders
                sortedPerformers = sorted(sortedPerformers, key = lambda x: x["title"])
                for performer_ in sortedPerformers:
                    performer = performer_["title"]
                    if Settings.get_category_performer():
                        maybeFiles = parse_categories(Settings.get_category_performer(), categoryFolder=performer_)
                        if len(maybeFiles) > 0: 
                            Settings.maybe_print("Performer: {}".format(performer))
                            return maybeFiles
                        continue
                    for cat_ in Settings.get_categories().remove("performers"):
                        maybeFiles = parse_categories(cat_, categoryFolder=performer_)
                        if len(maybeFiles) > 0: 
                            Settings.maybe_print("Performer: {}".format(performer))
                            return maybeFiles
            elif Settings.get_sort_method() == "random":
                randomPerformers = performerFolders
                random.shuffle(randomPerformers)
                randomCats = Settings.get_categories()
                randomCats.remove("performers")
                random.shuffle(randomCats)
                for performer_ in randomPerformers:
                    performer = performer_["title"]
                    if Settings.get_category_performer():
                        maybeFiles = parse_categories(Settings.get_category_performer(), categoryFolder=performer_)
                        if len(maybeFiles) > 0: 
                            Settings.maybe_print("Performer: {}".format(performer))
                            return maybeFiles
                        continue
                    for cat_ in randomCats:
                        maybeFiles = parse_categories(cat_, categoryFolder=performer_)
                        if len(maybeFiles) > 0: 
                            Settings.maybe_print("Performer: {}".format(performer))
                            return maybeFiles
        return parse_categories(cat)

    @staticmethod
    def get_random_file():
        files = Google_File.get_files()
        files = [file for file in files if not isinstance(file, str)]
        randomFile = random.choice(files)
        if Settings.get_sort_method() == "ordered":
            randomFile = sorted(files, key = lambda x: x.get_title())[0]
        Settings.maybe_print("Random File: {}".format(randomFile.get_title()))
        return randomFile

    def get_mimetype(self):
        if self.mimeType: return self.mimeType
        self.mimeType = self.get_file()["mimeType"]
        return self.mimeType

    def get_parent(self):
        if self.parent: return self.parent 
        self.parent = Google.get_file_parent(self.get_id())
        return self.parent

    def get_path(self):
        if self.path: return self.path
        # downloads to /tmp/downloads or whatever
        # if exists, adds 1 to end of name
        tmp = File.get_tmp()
        def counterfy():
            filename_ = str(self.get_title())+"{}"+str(self.get_ext())
            counter = 0
            while os.path.isfile(os.path.join(tmp, filename_.format(counter))):
                counter += 1
            filename_ = filename_.format(counter)
            Settings.maybe_print("filename: {}".format(filename_))
            filename_ = os.path.join(tmp, filename_.format(counter))
            return filename_
        filename = os.path.join(tmp, "{}{}".format(self.get_title(), self.get_ext()))
        if os.path.isfile(filename) and not Settings.is_prefer_local():
            filename = counterfy()
        # tmp = File.get_tmp() # i don't think this should be in file over settings
        filename = filename.strip('\'').strip('\"').strip()
        self.path = filename
        Settings.dev_print(self.path)
        return self.path

    def get_title(self):
        ## title would be set when created
        if self.title: return self.title
        title, ext = os.path.splitext(self.get_file()["title"])
        self.ext = ext
        self.title = title.replace(" ","_")
        return self.title

    # files are File references
    # file references can be GoogleId references which need to download their source
    # files exist when checked for size
    def prepare(self):
        Settings.maybe_print("preparing1: {}".format(self.get_title()))
        if not self.check_size():
            self.download()
        return super()

    @staticmethod
    def select_file(category, performer=None):
        if not Settings.is_prompt(): return Google_File.get_random_file()
        # this is a list of google files to select from
        files = Google_File.get_files_by_category(category, performer=performer)
        files_ = []
        for file in files:
            if isinstance(file, str):
                files_.append(PyInquirer.Separator())
                continue
            file.category = category
            file_ = {
                "name": file.file['title'],
                "value": file,
                "short": file.file['id']
            }
            files_.append(file_)
        if len(files_) == 0:
            print("Missing Files")
            return Google_File.select_files()
        question = {
            'type': 'list',
            'name': 'file',
            'message': 'Google Files:',
            'choices': files_,
            # 'filter': lambda file: file.lower()
        }
        file = PyInquirer.prompt(question)["file"]
        if not Settings.confirm(file.get_title()): return Google_File.select_file(category)
        return file

    @staticmethod
    def select_files():
        if not Settings.is_prompt(): return [Google_File.get_random_file()]
        category = Settings.select_category()
        if not category: return File.select_file_upload_method()
        # if not Settings.confirm(category): return Google_File.select_files()
        print("Select Google Files or a Folder")
        files = []
        while True:
            file = Google_File.select_file(category)
            if not file: break
            ##
            if "performer" in str(category):
                cat = Settings.select_category([cat for cat in Settings.get_categories() if "performer" not in cat])
                performerName = file.get_title()
                file = Google_File.select_file(cat, performer=performerName)
                if not file: break
                setattr(file, "performer", performerName)
                files.append(file)
                if "galler" in str(cat) or "video" in str(cat): break
            ##
            files.append(file)
            if "galler" in str(category) or "video" in str(category): break
        if not Settings.confirm([file.file['title'] for file in files]): return Google_File.select_files()
        return files

##########################################################################################

class Google_Folder(Google_File):
    def __init__(self):
        Google_File.__init__(self)
        self.files = None

    def backup(self):
        if File.backup_text(self.get_title()): return
        Google.upload_gallery(files=self.files)

    def check_size(self):
        for file in self.get_files():
            exists = file.check_size()
            if not exists: return False
        return True

    def download(self):
        print("Downloading Folder: {}".format(self.get_title()))
        if len(self.files) == 0:
            file_list = Google.get_files_by_folder_id(self.get_id())
            self.files = []
            for file in file_list:
                file_ = Google_File()
                setattr(file_, "id", file["id"])
                setattr(file_, "file", file)
                self.files.append(file_)
        folder_size = len(self.files)
        Settings.maybe_print('Folder size: {}'.format(folder_size))
        Settings.maybe_print('Upload limit: {}'.format(Settings.get_upload_max()))
        if int(folder_size) == 0:
            print('Error: Empty Folder')
            return False
        file_list = self.files
        random.shuffle(file_list)
        file_list = file_list[:int(Settings.get_upload_max())]
        ## video preference
        videos = []
        for file in file_list:
            if str(file.get_mimetype()) in MIMETYPES_VIDEOS_LIST:
                videos.append(file)
        if len(videos) > 0: file_list = [random.choice(videos)]
        ##
        i = 1
        for file in sorted(file_list, key = lambda x: x.get_title()):
            # print_same_line("Downloading: {} ({}/{})".format(file.get_title(), i, folder_size))
            print("Downloading: {} ({}/{})".format(file.get_title(), i, folder_size))
            file.download()
            i+=1
        print()
        print("Downloaded Folder: {}".format(self.get_title()))

    def get_files(self):
        if not self.files:
            self.files = []
            files = Google.get_files_by_folder_id(self.get_id())
            for file in files:
                file_ = Google_File()
                setattr(file_, "file", file)
                self.files.append(file_)
        if Settings.get_title():
            for file in self.files:
                if str(Settings.get_title()) == str(file.get_title()):
                    self.files = [file]
                    break
        return self.files

    def prepare():
        prepared = False
        for file in self.get_files():
            prepared_ = file.prepare()
            if prepared_: prepared = prepared_
        return prepared
        
###################################################################################

class Image(File):
    def __init__(self):
        pass

    def prepare(self):
        Settings.maybe_print("preparingi: {}".format(self.get_title()))
        return super()

###################################################################################

class Video(File):
    def __init__(self):
        self.screenshots = []
        self.trimmed = ""
        self.split = ""

    #seconds off front or back
    def trim(self):
        path = self.get_path()
        self.trimmed = ffmpeg.trim(path) 

    # into segments (60 sec, 5 min, 10 min)
    def split(self):
        path = self.get_path()
        self.split = ffmpeg.split(path)

    # unnecessary, handled by onlyfans
    # unless this somehow adds like more metadata
    def watermark(self):
        pass

    # cleanup & label appropriately (digital watermarking?)
    def get_metadata(self):
        pass

    # frames for preview gallery
    def get_frames(self):
        path = self.get_path()
        self.screenshots = ffmpeg.frames(path)

    def prepare(self):
        Settings.maybe_print("preparingv: {}".format(self.get_title()))
        self.reduce()
        self.repair()
        self.watermark()
        return super()

    def reduce(self):
        if not Settings.is_reduce(): 
            Settings.maybe_print("Skipping: Video Reduction")
            return
        path = self.get_path()
        global FIFTY_MEGABYTES
        if (int(os.stat(str(path)).st_size) < FIFTY_MEGABYTES or str(Settings.is_reduce()) == "False"):
            return
        Settings.dev_print("reduce: {}".format(self.get_title()))
        self.path = ffmpeg.reduce(path)
    
    # unnecessary
    def repair(self):
        if not Settings.is_repair():
            Settings.dev_print("Skipping: Video Repair")
            return
        path = self.get_path()
        if Settings.is_repair():
            return
        Settings.dev_print("repair: {}".format(self.get_title()))
        self.path = ffmpeg.repair(path)





def run_browserstack(self,os_name,os_version,browser,browser_version,remote_project_name,remote_build_name):
    "Run the test in browser stack when remote flag is 'Y'"
    #Get the browser stack credentials from browser stack credentials file
    USERNAME = remote_credentials.USERNAME
    PASSWORD = remote_credentials.ACCESS_KEY
    if browser.lower() == 'ff' or browser.lower() == 'firefox':
        desired_capabilities = DesiredCapabilities.FIREFOX            
    elif browser.lower() == 'ie':
        desired_capabilities = DesiredCapabilities.INTERNETEXPLORER
    elif browser.lower() == 'chrome':
        desired_capabilities = DesiredCapabilities.CHROME            
    elif browser.lower() == 'opera':
        desired_capabilities = DesiredCapabilities.OPERA        
    elif browser.lower() == 'safari':
        desired_capabilities = DesiredCapabilities.SAFARI
    desired_capabilities['os'] = os_name
    desired_capabilities['os_version'] = os_version
    desired_capabilities['browser_version'] = browser_version
    if remote_project_name is not None:
        desired_capabilities['project'] = remote_project_name
    if remote_build_name is not None:
        desired_capabilities['build'] = remote_build_name+"_"+str(datetime.now().strftime("%c"))

    return webdriver.Remote(RemoteConnection("http://%s:%s@hub-cloud.browserstack.com/wd/hub"%(USERNAME,PASSWORD),resolve_ip= False),
        desired_capabilities=desired_capabilities) 





def remote(self) -> None:
    if not self.debug:
        log("debug mode is turned off, can't reuse old session")
        return

    file = open(self.current_session_path, "r")
    content = file.read()
    lines = content.split(";")
    url = lines[0]
    session = lines[1]

    self.driver = webdriver.Remote(
        command_executor=url, desired_capabilities=DesiredCapabilities.CHROME
    )
    self.driver.session_id = session

    self.set_config() 




def attach_to_session(executor_url, session_id):
    original_execute = WebDriver.execute

    def new_command_execute(self, command, params=None):
        if command == "newSession":
            # Mock the response
            return {'success': 0, 'value': None, 'sessionId': session_id}
        else:
            return original_execute(self, command, params)

    # Patch the function before creating the driver object
    WebDriver.execute = new_command_execute
    driver = webdriver.Remote(command_executor=executor_url,
                              desired_capabilities={})
    driver.session_id = session_id
    # Replace the patched function with original function
    WebDriver.execute = original_execute
    return driver 




def __init__(self, command_executor = None, session_id = None, previous_read_message = None):
    chrome_options = Options()
    chrome_options.add_argument("--no-sandbox")

    chrome_options.add_argument("--user-agent='Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.50 Safari/537.36'")

    chrome_options.add_argument('--verbose')
    chrome_options.add_argument('--log-path=/tmp/chrome.log')


    if command_executor == None:
        if config.CHROMEDRIVER_LOCATION == 'None': #Launching in docker container
            chrome_options.add_argument('--headless')
            self.driver = webdriver.Chrome(chrome_options=chrome_options)
            self.driver.set_window_size(1920, 1080)

        else:
            self.driver = webdriver.Chrome(config.CHROMEDRIVER_LOCATION, chrome_options=chrome_options)

        self.driver.get('https://web.whatsapp.com')

        print(self.driver.session_id )
        print(self.driver.command_executor._url)

        self.find_scan_code()

    else:
        self.driver = webdriver.Remote(command_executor=command_executor,desired_capabilities={})
        self.driver.close()
        self.driver.session_id = session_id

    self.previous_read_message = previous_read_message 







from selenium import webdriver

driver = webdriver.Chrome()
executor_url = driver.command_executor._url
session_id = driver.session_id
driver.get("http://tarunlalwani.com")

print session_id
print executor_url


driver2 = webdriver.Remote(command_executor=executor_url, desired_capabilities={})
driver2.session_id = session_id
print driver2.current_url








driver = webdriver.Firefox()  #python

# extract to session_id and _url from driver object.

url = driver.command_executor._url       #"http://127.0.0.1:60622/hub"
session_id = driver.session_id            #'4e167f26-dc1d-4f51-a207-f761eaf73c31'

# Use these two parameter to connect to your driver.

driver = webdriver.Remote(command_executor=url,desired_capabilities={})
driver.close()   # this prevents the dummy browser
driver.session_id = session_id

# And you are connected to your driver again.

driver.get("http://www.mrsmart.in")











add `destination` to backup functions

-keep -> keep browser open
-reconnect -> reconnect to existing/previous session

check if present @ spawn, determines reconnect behaviour without extra args
-session $sessionId -> session to reconnect to
-url $executorURL -> executor_url in conjunct w/ session

if keep: save session&url in mount_path
if reconnect & no session||url: check for saved session||url in mount_path



from .classes import Discount, Poll, Promotion, Schedule
from .driver import Driver
from .settings import Settings
from .user import User
from .profile import Profile
from .snarf import Snarf
from .colorize import colorize
from .file import File, Google_File, Google_Folder
from .validators import NumberValidator, TimeValidator, DateValidator, DurationValidator, ExpirationValidator, ListValidator
# from .google import google as Google
from . import args